import pytest
from django.urls import reverse
from django.test import Client
from camomilla.models.page import Page

@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_sitemap_xml_contains_pages():
    Page.objects.create(title="Test Page 1", permalink="test-page-1", status="PUB", autopermalink=False)
    Page.objects.create(title="Test Page 2", permalink="test-page-2", status="PUB", autopermalink=False)

    client = Client()
    response = client.get(reverse('django.contrib.sitemaps.views.sitemap'))
    assert response.status_code == 200
    assert b"<urlset" in response.content
    assert b"/test-page-1" in response.content
    assert b"/test-page-2" in response.content



import pytest
from django.test import TransactionTestCase
from rest_framework.test import APIClient
from .utils.api import login_superuser
from camomilla.models import Page
from camomilla.models.page import UrlRedirect
from datetime import datetime, timedelta


class PagesTestCase(TransactionTestCase):
    reset_sequences = True
    
    def setUp(self):
        self.client = APIClient()
        token = login_superuser()
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)

    def test_pages_api_no_access(self):
        client = APIClient()
        response = client.post("/api/camomilla/pages/")
        assert response.status_code == 401

    def test_pages_api_crud(self):
        # Create page
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "translations": {
                    "it": {
                        "title": "title_page_1",
                    }
                }
            },
            format="json",
        )

        assert response.status_code == 201
        assert len(Page.objects.all()) == 1
        page = Page.objects.first()
        assert page.id == 1
        assert page.title_it == "title_page_1"
        assert page.url_node.id == 1

        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "title_it": "title_page_2",
            },
        )

        assert response.status_code == 201
        assert len(Page.objects.all()) == 2
        page = Page.objects.last()
        assert page.id == 2
        assert page.title_it == "title_page_2"
        assert page.url_node.id == 2

        # Update page
        response = self.client.patch(
            "/api/camomilla/pages/2/",
            {
                "translations": {
                    "it": {
                        "title": "title_page_2_updated",
                    }
                }
            },
            format="json",
        )

        assert response.status_code == 200
        assert len(Page.objects.all()) == 2
        page = Page.objects.last()
        assert page.id == 2
        assert page.title_it == "title_page_2_updated"

        # Read page
        response = self.client.get("/api/camomilla/pages/2/")

        assert response.status_code == 200
        assert response.json()["id"] == 2
        assert response.json()["title"] == "title_page_2_updated"

        # Read pages
        response = self.client.get("/api/camomilla/pages/")

        assert response.status_code == 200
        assert response.json()[0]["id"] == 1
        assert response.json()[0]["title"] == "title_page_1"
        assert response.json()[1]["id"] == 2
        assert response.json()[1]["title"] == "title_page_2_updated"

        # Delete page
        response = self.client.delete("/api/camomilla/pages/2/")

        assert response.status_code == 204
        assert len(Page.objects.all()) == 1
        page = Page.objects.last()
        assert page.id == 1
        assert page.title_it == "title_page_1"

    def test_pages_url_nodes(self):
        # Create page with automatic url creation
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "title_en": "title_page_1",
                "title_it": "titolo_pagina_1",
            },
        )

        assert response.status_code == 201

        # EN automatic url creation
        response = self.client.get("/api/camomilla/pages/1/?language=en")
        assert response.json()["autopermalink"] == True
        assert response.json()["permalink"] == "/title_page_1"
        # IT automatic url creation
        response = self.client.get("/api/camomilla/pages/1/?language=it")
        assert response.json()["autopermalink"] == True
        assert response.json()["permalink"] == "/titolo_pagina_1"

        # Create page with manual url creation
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "translations": {
                    "it": {
                        "title": "titolo_pagina_2",
                        "permalink": "permalink_manuale_it_2",
                        "autopermalink": False,
                    },
                    "en": {
                        "title": "title_page_2",
                        "permalink": "permalink_manual_en_2",
                        "autopermalink": False,
                    },
                }
            },
            format="json",
        )
        assert response.status_code == 201

        # EN manual url creation
        response = self.client.get("/api/camomilla/pages/2/?language=en")
        assert response.json()["autopermalink"] == False
        assert response.json()["permalink"] == "/permalink_manual_en_2"
        # IT manual url creation
        response = self.client.get("/api/camomilla/pages/2/?language=it")
        assert response.json()["autopermalink"] == False
        assert response.json()["permalink"] == "/permalink_manuale_it_2"

        # Create page with a parent page with automatic url creation
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "title_en": "title_page_3",
                "title_it": "titolo_pagina_3",
                "parent_page": 2,
            },
        )
        assert response.status_code == 201

        # EN parent page with automatic url creation
        response = self.client.get("/api/camomilla/pages/3/?language=en")
        assert response.json()["autopermalink"] == True
        assert response.json()["permalink"] == "/permalink_manual_en_2/title_page_3"
        # IT parent page with automatic url creation
        response = self.client.get("/api/camomilla/pages/3/?language=it")
        assert response.json()["autopermalink"] == True
        assert response.json()["permalink"] == "/permalink_manuale_it_2/titolo_pagina_3"

        # Check url uniqueness and consistency EN
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "autopermalink_en": False,
                "permalink_en": "permalink_manual_en_2",
            },
        )

        # Client error when url check uniqueness and consistency fail
        assert response.status_code == 400
        assert (
            response.data["permalink_en"][0]
            == "There is an other page with same permalink."
        )

        # Check url uniqueness and consistency IT
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "translations": {
                    "it": {
                        "autopermalink": False,
                        "permalink": "permalink_manuale_it_2",
                    }
                }
            },
            format="json",
        )

        # Client error when url check uniqueness and consistency fail
        assert response.status_code == 400
        assert (
            response.data["permalink_it"][0]
            == "There is an other page with same permalink."
        )

    def test_pages_url_nodes_navigation(self):
        # Test the camomilla.dynamic_pages_url handler for navigating and rendering UrlNodes
        self.client.post(
            "/api/camomilla/pages/",
            {
                "autopermalink_en": False,
                "permalink_en": "permalink_4_en",
                "status_en": "PUB",
                "autopermalink_it": False,
                "permalink_it": "permalink_4_it",
                "status_it": "PUB",
            },
        )

        response = self.client.get("/permalink_4_en/")
        assert response.status_code == 200
        response = self.client.get("/it/permalink_4_it/")
        assert response.status_code == 200

        # Test draft - published - planned and ?preview=true
        self.client.post(
            "/api/camomilla/pages/",
            {
                "translations": {
                    "it": {
                        "autopermalink": False,
                        "permalink": "permalink_5_it",
                        "status": "PLA",
                    },
                    "en": {
                        "autopermalink": False,
                        "permalink": "permalink_5_en",
                        "status": "DRF",
                    },
                }
            },
            format="json",
        )

        response = self.client.get("/permalink_5_en/")
        assert response.status_code == 404
        response = self.client.get("/permalink_5_en/?preview=true")
        assert response.status_code == 200
        response = self.client.get("/it/permalink_5_it/")
        assert response.status_code == 404

        self.client.patch(
            "/api/camomilla/pages/2/",
            {
                "publication_date": (datetime.now() - timedelta(1)).strftime("%Y-%m-%d")
                + " 00:00:00",
            },
        )

        response = self.client.get("/it/permalink_5_it/")
        assert response.status_code == 200

    def test_pages_url_nodes_navigation_redirects(self):
        # Test the camomilla.dynamic_pages_url handler for navigating and rendering UrlNodes
        self.client.post(
            "/api/camomilla/pages/",
            {
                "translations": {
                    "it": {
                        "autopermalink": False,
                        "permalink": "permalink_6_it",
                        "status": "PUB",
                    },
                    "en": {
                        "autopermalink": False,
                        "permalink": "permalink_6_en",
                        "status": "PUB",
                    },
                }
            },
            format="json",
        )

        # EN Insert Moved Permanently Redirect
        url_redirect = UrlRedirect.objects.create(
            language_code="en",
            from_url="/redirecting_1",
            to_url="/redirected_1",
            url_node_id=1,
        )

        response = self.client.get("/redirecting_1/")
        assert response.status_code == 301
        assert response.url == "/redirected_1/"

        # EN Change to Moved Temporarily Redirect
        url_redirect.permanent = False
        url_redirect.save()
        response = self.client.get("/redirecting_1/")
        assert response.status_code == 302

        # IT Insert Moved Permanently Redirect
        url_redirect = UrlRedirect.objects.create(
            language_code="it",
            from_url="/urlreindirizzamento_1",
            to_url="/urlreindirizzato_1",
            url_node_id=1,
        )

        response = self.client.get("/it/urlreindirizzamento_1/")
        assert response.status_code == 301
        assert response.url == "/it/urlreindirizzato_1/"

        # IT Change to Moved Temporarily Redirect
        url_redirect.permanent = False
        url_redirect.save()
        response = self.client.get("/it/urlreindirizzamento_1/")
        assert response.status_code == 302

        # Test auto redirect after permalink change
        self.client.patch(
            "/api/camomilla/pages/1/",
            {
                "translations": {
                    "it": {
                        "permalink": "permalink_6_it_changed",
                    },
                    "en": {
                        "permalink": "permalink_6_en_changed",
                    },
                }
            },
            format="json",
        )

        response = self.client.get("/permalink_6_en/")
        assert response.status_code == 301
        assert response.url == "/permalink_6_en_changed/"

        response = self.client.get("/it/permalink_6_it/")
        assert response.status_code == 301
        assert response.url == "/it/permalink_6_it_changed/"

    def test_page_keywords(self):
        # Create page with keywords field and check it's given back as expected
        response = self.client.post(
            "/api/camomilla/pages/",
            {
                "og_description_it" : "Keywords Test",
                "keywords_it" : 'key1, key2'
            },
            format="json",
        )

        assert response.status_code == 201
        assert len(Page.objects.all()) == 1
        page = Page.objects.first()
        assert page.og_description_it == "Keywords Test"
        assert page.keywords_it == 'key1, key2'


from typing import Sequence
from .getters import pointed_getter


def set_key(data, key, val):
    if isinstance(data, Sequence):
        key = int(key)
        if key < len(data):
            data[key] = val
        else:
            data.append(val)
        return data
    elif isinstance(data, dict):
        data[key] = val
    else:
        setattr(data, key, val)
    return data


def get_key(data, key, default):
    if isinstance(data, Sequence):
        try:
            return data[int(key)]
        except IndexError:
            return default
    return pointed_getter(data, key, default)


def pointed_setter(data, path, value):
    path = path.split(".")
    key = path.pop(0)
    if not len(path):
        return set_key(data, key, value)
    default = [] if path[0].isdigit() else {}
    return set_key(
        data, key, pointed_setter(get_key(data, key, default), ".".join(path), value)
    )



from typing import Sequence, Tuple, Optional, Union
from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist

from django.db import models, transaction
from django.db.models.signals import post_delete, post_save, pre_save
from django.dispatch import receiver
from django.http import Http404, HttpRequest
from django.shortcuts import redirect
from django.urls import NoReverseMatch, reverse
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from django.utils.translation import get_language

from camomilla.managers.pages import PageQuerySet, UrlNodeManager
from camomilla.models.mixins import MetaMixin, SeoMixin
from camomilla.utils import (
    activate_languages,
    get_field_translations,
    get_nofallbacks,
    lang_fallback_query,
    set_nofallbacks,
    url_lang_decompose,
)
from camomilla.utils.getters import pointed_getter
from camomilla import settings
from camomilla.templates_context.rendering import ctx_registry
from django.conf import settings as django_settings
from modeltranslation.utils import build_localized_fieldname
from django.utils.module_loading import import_string


class UrlRedirect(models.Model):
    language_code = models.CharField(max_length=10, null=True)
    from_url = models.CharField(max_length=400)
    to_url = models.CharField(max_length=400)
    url_node = models.ForeignKey(
        "UrlNode", on_delete=models.CASCADE, related_name="redirects"
    )
    permanent = models.BooleanField(default=True)
    date_created = models.DateTimeField(auto_now_add=True)
    date_updated_at = models.DateTimeField(auto_now=True)

    __q_string = ""

    def __str__(self) -> str:
        return f"[{self.language_code}] {self.from_url} -> {self.to_url}"

    @classmethod
    def find_redirect(
        cls, request: HttpRequest, language_code: Optional[str] = None
    ) -> Optional["UrlRedirect"]:
        instance = cls.find_redirect_from_url(request.path, language_code)
        if instance:
            instance.__q_string = request.META.get("QUERY_STRING", "")
        return instance

    @classmethod
    def find_redirect_from_url(
        cls, from_url: str, language_code: Optional[str] = None
    ) -> Optional["UrlRedirect"]:
        path_decomposition = url_lang_decompose(from_url)
        language_code = (
            language_code or path_decomposition["language"] or get_language()
        )
        from_url = path_decomposition["permalink"]
        return cls.objects.filter(
            from_url=from_url.rstrip("/"), language_code=language_code or get_language()
        ).first()

    def redirect(self) -> str:
        return redirect(self.redirect_to, permanent=self.permanent)

    @property
    def redirect_to(self) -> str:
        url_to = "/" + self.to_url.lstrip("/")
        if getattr(django_settings, "APPEND_SLASH", True) and not url_to.endswith("/"):
            url_to += "/"
        if (
            self.language_code != settings.DEFAULT_LANGUAGE
            and settings.ENABLE_TRANSLATIONS
        ):
            url_to = "/" + self.language_code + url_to
        return url_to + ("?" + self.__q_string if self.__q_string else "")

    class Meta:
        verbose_name = _("Redirect")
        verbose_name_plural = _("Redirects")
        unique_together = ("from_url", "language_code")
        indexes = [
            models.Index(fields=["from_url", "language_code"]),
        ]


class UrlNode(models.Model):

    LANG_PERMALINK_FIELDS = (
        [
            build_localized_fieldname("permalink", lang)
            for lang in settings.LANGUAGE_CODES
        ]
        if settings.ENABLE_TRANSLATIONS
        else ["permalink"]
    )

    permalink = models.CharField(max_length=400, unique=True, null=True)
    related_name = models.CharField(max_length=200)
    objects = UrlNodeManager()

    @property
    def page(self) -> "AbstractPage":
        return getattr(self, self.related_name)

    @staticmethod
    def reverse_url(permalink: str, request: Optional[HttpRequest] = None) -> str:
        append_slash = getattr(django_settings, "APPEND_SLASH", True)
        try:
            if permalink == "/":
                return reverse("camomilla-homepage")
            url = reverse("camomilla-permalink", args=(permalink.lstrip("/"),))
            if append_slash and not url.endswith("/"):
                url += "/"
            if request:
                url = request.build_absolute_uri(url)
            return url
        except NoReverseMatch:
            return None

    @property
    def routerlink(self) -> str:
        return self.reverse_url(self.permalink) or self.permalink

    def get_absolute_url(self) -> str:
        if self.routerlink == "/":
            return ""
        return self.routerlink

    @staticmethod
    def sanitize_permalink(permalink):
        if isinstance(permalink, str):
            p_parts = permalink.split("/")
            permalink = "/".join(
                [slugify(p, allow_unicode=True).strip() for p in p_parts]
            )
            if not permalink.startswith("/"):
                permalink = f"/{permalink}"
        return permalink

    def save(self, *args, **kwargs) -> None:
        for lang_p_field in UrlNode.LANG_PERMALINK_FIELDS:
            setattr(
                self,
                lang_p_field,
                UrlNode.sanitize_permalink(getattr(self, lang_p_field)),
            )
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return self.permalink


PAGE_CHILD_RELATED_NAME = "%(app_label)s_%(class)s_child_pages"
URL_NODE_RELATED_NAME = "%(app_label)s_%(class)s"

PAGE_STATUS = (
    ("PUB", _("Published")),
    ("DRF", _("Draft")),
    ("TRS", _("Trash")),
    ("PLA", _("Planned")),
)


class PageBase(models.base.ModelBase):
    """
    This models comes to implement a language based permalink logic
    """

    def perm_prop_factory(permalink_field):
        def getter(_self):
            return getattr(
                _self,
                f"__{permalink_field}",
                getattr(_self.url_node or object(), permalink_field, None),
            )

        def setter(_self, value: str):
            setattr(_self, f"__{permalink_field}", value)

        return getter, setter

    def __new__(cls, name, bases, attrs, **kwargs):
        attr_meta = attrs.pop("PageMeta", None)
        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        page_meta = attr_meta or getattr(new_class, "PageMeta", None)
        base_page_meta = getattr(new_class, "_page_meta", None)
        for lang_p_field in UrlNode.LANG_PERMALINK_FIELDS:
            computed_prop = property(*cls.perm_prop_factory(lang_p_field))
            setattr(new_class, lang_p_field, computed_prop)
        if settings.ENABLE_TRANSLATIONS:
            setattr(
                new_class,
                "permalink",
                property(
                    lambda _self: getattr(
                        _self,
                        build_localized_fieldname("permalink", get_language()),
                        None,
                    ),
                    lambda _self, value: setattr(
                        _self,
                        f"__{build_localized_fieldname('permalink', get_language())}",
                        value,
                    ),
                ),
            )
        if page_meta:
            for name, value in getattr(base_page_meta, "__dict__", {}).items():
                if name not in page_meta.__dict__:
                    setattr(page_meta, name, value)
            setattr(new_class, "_page_meta", page_meta)
        return new_class


class AbstractPage(SeoMixin, MetaMixin, models.Model, metaclass=PageBase):
    identifier = models.CharField(max_length=200, null=True, unique=True, default=uuid4)
    date_created = models.DateTimeField(auto_now_add=True)
    date_updated_at = models.DateTimeField(auto_now=True)
    breadcrumbs_title = models.CharField(max_length=128, null=True, blank=True)
    template = models.CharField(max_length=500, null=True, blank=True)
    template_data = models.JSONField(default=dict, null=False, blank=True)
    ordering = models.PositiveIntegerField(default=0, blank=False, null=False)
    parent_page = models.ForeignKey(
        "self",
        related_name=PAGE_CHILD_RELATED_NAME,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    url_node = models.OneToOneField(
        UrlNode,
        on_delete=models.CASCADE,
        related_name=URL_NODE_RELATED_NAME,
        null=True,
        editable=False,
    )
    publication_date = models.DateTimeField(null=True, blank=True)
    status = models.CharField(
        max_length=3,
        choices=PAGE_STATUS,
        default="DRF",
    )
    indexable = models.BooleanField(default=True)
    autopermalink = models.BooleanField(default=True)

    objects = PageQuerySet.as_manager()

    __cached_db_instance: "AbstractPage" = None

    @property
    def db_instance(self):
        if self.__cached_db_instance is None:
            self.__cached_db_instance = self.get_db_instance()
        return self.__cached_db_instance

    def get_db_instance(self):
        if self.pk:
            return self.__class__.objects.get(pk=self.pk)
        return None

    def __init__(self, *args, **kwargs):
        super(AbstractPage, self).__init__(*args, **kwargs)

    def __str__(self) -> str:
        return "(%s) %s" % (self.__class__.__name__, self.title or self.permalink)

    def get_context(self, request: Optional[HttpRequest] = None):
        context = {
            "page": self,
            "page_model": {"class": self.__class__.__name__, "module": self.__module__},
            "request": request,
        }
        inject_func = pointed_getter(self, "_page_meta.inject_context_func")
        if inject_func and callable(inject_func):
            new_ctx = inject_func(request=request, super_ctx=context)
            if isinstance(new_ctx, dict):
                context.update(new_ctx)
        return ctx_registry.get_context_for_page(self, request, super_ctx=context)

    @classmethod
    def get_serializer(cls):
        from camomilla.serializers.mixins import AbstractPageMixin

        standard_serializer = (
            pointed_getter(cls, "_page_meta.standard_serializer")
            or settings.PAGES_DEFAULT_SERIALIZER
        )
        if isinstance(standard_serializer, str):
            standard_serializer = import_string(standard_serializer)
        if not issubclass(standard_serializer, AbstractPageMixin):
            raise ValueError(
                f"Standard serializer {standard_serializer} must be a subclass of AbstractPageMixin"
            )
        return standard_serializer

    @property
    def model_name(self) -> str:
        return self._meta.app_label + "." + self._meta.model_name

    @property
    def model_info(self) -> dict:
        return {"app_label": self._meta.app_label, "class": self._meta.model_name}

    @property
    def routerlink(self) -> str:
        return self.url_node and self.url_node.routerlink

    @property
    def breadcrumbs(self) -> Sequence[dict]:
        breadcrumb = {
            "permalink": self.routerlink,
            "title": self.breadcrumbs_title or self.title or "",
        }
        if self.parent:
            return self.parent.breadcrumbs + [breadcrumb]
        return [breadcrumb]

    @property
    def is_public(self) -> bool:
        status = get_nofallbacks(self, "status")
        publication_date = get_nofallbacks(self, "publication_date")
        if status == "PUB":
            return True
        if status == "PLA":
            return bool(publication_date) and timezone.now() > publication_date
        return False

    def get_template_path(self, request: Optional[HttpRequest] = None) -> str:
        return self.template or pointed_getter(self, "_page_meta.default_template")

    @property
    def childs(self) -> models.Manager:
        if hasattr(self._page_meta, "child_page_field"):
            return getattr(self, self._page_meta.child_page_field)
        return getattr(self, PAGE_CHILD_RELATED_NAME % self.model_info, self.__class__.objects.none())

    @property
    def parent(self) -> models.Model:
        return getattr(self, self._page_meta.parent_page_field)

    def _get_or_create_url_node(self) -> UrlNode:
        if not self.url_node:
            self.url_node = UrlNode.objects.create(
                related_name=URL_NODE_RELATED_NAME % self.model_info
            )
        return self.url_node

    def _update_url_node(self, force: bool = False) -> UrlNode:
        self.url_node = self._get_or_create_url_node()
        for __ in activate_languages():
            old_permalink = self.db_instance and self.db_instance.permalink
            new_permalink = self.permalink
            if self.autopermalink:
                new_permalink = self.generate_permalink()
            force = force or old_permalink != new_permalink
            set_nofallbacks(self.url_node, "permalink", new_permalink)
        if force:
            self.url_node.save()
            self.update_childs()
        return self.url_node

    def generate_permalink(self, safe: bool = True) -> str:
        permalink = f"/{slugify(self.title or '', allow_unicode=True)}"
        if self.parent:
            parent_permalink = (self.parent.permalink or "").lstrip("/")
            permalink = f"/{parent_permalink}{permalink}"
        set_nofallbacks(self, "permalink", permalink)
        qs = UrlNode.objects.exclude(pk=getattr(self.url_node or object, "pk", None))
        if safe and qs.filter(permalink=permalink).exists():
            permalink = "/".join(
                permalink.split("/")[:-1] + [slugify(uuid4(), allow_unicode=True)]
            )
        return permalink

    def update_childs(self) -> None:
        # without pk, no childs there
        if self.pk is not None:
            exclude_kwargs = {}
            if self.childs.model == self.__class__:
                exclude_kwargs["pk"] = self.pk
            for child in self.childs.exclude(**exclude_kwargs):
                child.save()

    def save(self, *args, **kwargs) -> None:
        with transaction.atomic():
            self._update_url_node()
            super().save(*args, **kwargs)
            self.__cached_db_instance = None
            for lang_p_field in UrlNode.LANG_PERMALINK_FIELDS:
                hasattr(self, f"__{lang_p_field}") and delattr(
                    self, f"__{lang_p_field}"
                )

    @classmethod
    def get(cls, request: HttpRequest, *args, **kwargs) -> "AbstractPage":
        bypass_type_check = kwargs.pop("bypass_type_check", False)
        bypass_public_check = kwargs.pop("bypass_public_check", False)
        if len(kwargs.keys()) > 0:
            page = cls.objects.get(**kwargs)
        else:
            if not request:
                raise ValueError("request is required if no kwargs are passed")
            path = request.path
            if getattr(django_settings, "APPEND_SLASH", True):
                path = path.rstrip("/")
            node = UrlNode.objects.filter(
                permalink=url_lang_decompose(path)["permalink"]
            ).first()
            page = node and node.page
        type_error = not bypass_type_check and not isinstance(page, cls)
        public_error = not bypass_public_check and not getattr(
            page or object, "is_public", False
        )
        if not page or type_error or public_error:
            bases = (UrlNode.DoesNotExist,)
            if hasattr(cls, "DoesNotExist"):
                bases += (cls.DoesNotExist,)
            message = "%s matching query does not exist." % cls._meta.object_name
            if public_error:
                message = (
                    "Match found: %s.\nThe page appears not to be public.\nUse ?preview=true in the url to see it."
                    % page
                )
            raise type("PageDoesNotExist", bases, {})(message)
        return page

    @classmethod
    def get_or_create(
        cls, request: HttpRequest, *args, **kwargs
    ) -> Tuple["AbstractPage", bool]:
        try:
            return cls.get(request, *args, **kwargs), False
        except ObjectDoesNotExist:
            if len(kwargs.keys()) > 0:
                return cls.objects.get_or_create(**kwargs)
        return (None, False)

    @classmethod
    def get_or_create_homepage(cls) -> Tuple["AbstractPage", bool]:
        try:
            if settings.ENABLE_TRANSLATIONS:
                node = UrlNode.objects.get(lang_fallback_query(permalink="/"))
            else:
                node = UrlNode.objects.get(permalink="/")
            return node.page, False
        except UrlNode.DoesNotExist:
            return cls.get_or_create(None, permalink="/")

    @classmethod
    def get_or_404(cls, request: HttpRequest, *args, **kwargs) -> "AbstractPage":
        try:
            return cls.get(request, *args, **kwargs)
        except ObjectDoesNotExist as ex:
            raise Http404(ex)

    def alternate_urls(self, *args, **kwargs) -> dict:
        request: Union[HttpRequest, bool] = False
        if len(args) > 0:
            request = args[0]
        if "request" in kwargs:
            request = kwargs["request"]
        preview = request and getattr(request, "GET", {}).get("preview", False)
        permalinks = get_field_translations(self.url_node or object, "permalink", None)
        for lang in activate_languages():
            if lang in permalinks and permalinks[lang]:
                permalinks[lang] = (
                    UrlNode.reverse_url(permalinks[lang])
                    if preview or self.is_public
                    else None
                )
            if preview:
                permalinks = {k: f"{v}?preview=true" for k, v in permalinks.items()}
        permalinks.pop(get_language(), None)
        return permalinks

    class Meta:
        abstract = True
        ordering = ("ordering",)
        verbose_name = _("Page")
        verbose_name_plural = _("Pages")

    class PageMeta:
        parent_page_field = "parent_page"
        default_template = settings.PAGE_DEFAULT_TEMPLATE
        inject_context_func = settings.PAGE_INJECT_CONTEXT_FUNC
        standard_serializer = settings.PAGES_DEFAULT_SERIALIZER


class Page(AbstractPage):
    pass


@receiver(post_delete)
def auto_delete_url_node(sender, instance, **kwargs):
    if issubclass(sender, AbstractPage):
        instance.url_node and instance.url_node.delete()


__url_node_history__ = {}


@receiver(pre_save, sender=UrlNode)
def cache_url_node(sender, instance, **kwargs):
    if instance.pk:
        __url_node_history__[instance.pk] = sender.objects.filter(
            pk=instance.pk
        ).first()


@receiver(post_save, sender=UrlNode)
def generate_redirects(sender, instance, **kwargs):
    previous = __url_node_history__.pop(instance.pk, None)
    if previous:
        redirects = []
        with transaction.atomic():
            for lang in activate_languages():
                new_permalink = get_nofallbacks(instance, "permalink")
                old_permalink = get_nofallbacks(previous, "permalink")
                UrlRedirect.objects.filter(
                    from_url=new_permalink, language_code=lang
                ).delete()
                if old_permalink and old_permalink != new_permalink:
                    redirects.append(
                        UrlRedirect(
                            from_url=old_permalink,
                            to_url=new_permalink,
                            url_node=instance,
                            language_code=lang,
                        )
                    )
                    UrlRedirect.objects.filter(
                        to_url=old_permalink, language_code=lang
                    ).update(to_url=new_permalink)
            if len(redirects) > 0:
                UrlRedirect.objects.bulk_create(redirects)



import pytest
from django.test import TestCase
from rest_framework.test import APIClient

from tests.utils.media import load_asset_and_remove_media
from .utils.api import login_superuser
from camomilla.models import Page, Media
from example.website.models import CustomPageMetaModel, InvalidPageMetaModel


class PagaMetaTestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        token = login_superuser()
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)

    @pytest.mark.django_db(transaction=True, reset_sequences=True)
    def test_page_meta_rendering(self):
        asset = load_asset_and_remove_media("37059501.png")
        Media.objects.create(
            file=asset,
            alt_text="Test Media",
            title="Test Media",
            description="Description of test media",
        )
        page = CustomPageMetaModel.objects.create(
            title="Test Page",
            custom_field="Custom Data",
            permalink="test-page",
            status="PUB",
            autopermalink=False,
        )
        page.save()
        response = self.client.get("/test-page/")
        assert response.status_code == 200
        content = response.content.decode()
        assert "👻 I&#x27;m beeing injected!" in content
        assert "<h1>I'm the custom template!</h1>" in content
        assert "<ul><li>Test Media</li></ul>" in content

    @pytest.mark.django_db(transaction=True, reset_sequences=True)
    def test_page_meta_custom_parent_page(self):
        parent_page = Page.objects.create(
            title="Parent Page",
            permalink="parent-page",
            status="PUB",
            autopermalink=False,
        )
        child_page = CustomPageMetaModel.objects.create(
            title="Child Page",
            custom_field="Child Data",
            status="PUB",
            custom_parent_page=parent_page,
        )
        assert child_page.permalink == "/parent-page/child-page"

    @pytest.mark.django_db(transaction=True, reset_sequences=True)
    def test_page_meta_custom_serializer(self):

        CustomPageMetaModel.objects.create(
            title="Test Page with Custom Serializer",
            custom_field="Custom Data",
            permalink="test-page-custom-serializer",
            status="PUB",
            autopermalink=False,
        )

        response = self.client.get(
            "/api/camomilla/pages-router/test-page-custom-serializer"
        )
        assert response.status_code == 200
        data = response.json()
        assert data["title"] == "Test Page with Custom Serializer"
        assert data["custom_field"] == "Custom Data"
        assert data["permalink"] == "/test-page-custom-serializer"
        assert (
            data["serializer_custom_field"]
            == "I'm coming from CustomPageSerializer! 🫡"
        )

    @pytest.mark.django_db(transaction=True, reset_sequences=True)
    def test_page_meta_custom_serializer_error(self):
        with pytest.raises(ValueError) as exc_info:
            InvalidPageMetaModel.get_serializer()
            assert (
                str(exc_info.value)
                == "Standard serializer <class 'example.website.serializers.InvalidSerializer'> must be a subclass of AbstractPageMixin"
            )



import json

from django.db import connection

from django.db.models import JSONField as DjangoJSONField
from django.db import models


if connection.vendor == "sqlite":  # noqa: C901

    class JSONField(models.Field):
        def db_type(self, connection):
            return "text"

        def from_db_value(self, value, expression, connection):
            if value is not None:
                return self.to_python(value)
            return value

        def to_python(self, value):
            if value is not None:
                try:
                    return json.loads(value)
                except (TypeError, ValueError):
                    return value
            return value

        def get_prep_value(self, value):
            if value is not None:
                return str(json.dumps(value))
            return value

        def value_to_string(self, obj):
            return self.value_from_object(obj)

else:

    class JSONField(DjangoJSONField):
        pass


__all__ = [JSONField]



from django.db.models.aggregates import Max
from django.db.models.functions import Coalesce
from camomilla.fields import ORDERING_ACCEPTED_FIELDS


class OrderingMixin:
    """
    This mixin allows to set the default value of an ordering field to the max value + 1.
    """

    def get_max_order(self, order_field):
        return self.Meta.model.objects.aggregate(
            max_order=Coalesce(Max(order_field), 0)
        )["max_order"]

    def _get_ordering_field_name(self):
        try:
            field_name = self.Meta.model._meta.ordering[0]
            if field_name[0] == "-":
                field_name = field_name[1:]
            return field_name
        except (AttributeError, IndexError):
            return None

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )
        if (
            isinstance(model_field, ORDERING_ACCEPTED_FIELDS)
            and field_name == self._get_ordering_field_name()
        ):
            field_kwargs["default"] = self.get_max_order(field_name) + 1
        return field_class, field_kwargs



from camomilla.utils import get_host_url, get_complete_url, get_templates

from django.test import TestCase
from django.test import RequestFactory
import responses
from camomilla.settings import INTEGRATIONS_ASTRO_URL


class UtilsTestCase(TestCase):
    def setUp(self):
        self.astro_api_url = INTEGRATIONS_ASTRO_URL + "/api/templates"

    def test_get_host_url(self):
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        host_url = get_host_url(request)
        self.assertEqual(host_url, "http://localhost")
        host_url = get_host_url(None)
        self.assertEqual(host_url, None)

    def test_get_complete_url(self):
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        complete_url = get_complete_url(request, "path")
        self.assertEqual(complete_url, "http://localhost/path")
        complete_url = get_complete_url(request, "path", "it")
        self.assertEqual(complete_url, "http://localhost/it/path")
        complete_url = get_complete_url(request, "path", "fr")
        self.assertEqual(complete_url, "http://localhost/fr/path")

    @responses.activate
    def test_get_all_templates_files_error(self):
        responses.add(
            responses.GET,
            self.astro_api_url,
            json=["Error"],
            status=400,
        )
        templates = get_templates(request=RequestFactory().get("/"))
        self.assertFalse("Astro: Error" in templates)
        self.assertEqual(responses.calls[0].request.url, self.astro_api_url)

    @responses.activate
    def test_get_all_templates_files(self):
        responses.add(
            responses.GET,
            self.astro_api_url,
            json=["mock_template/1", "mock_template/2"],
            status=200,
        )
        templates = get_templates(request=RequestFactory().get("/"))
        self.assertTrue("mock_template/1" in templates)
        self.assertTrue("mock_template/2" in templates)
        self.assertEqual(responses.calls[0].request.url, self.astro_api_url)



from pathlib import Path
from typing import Sequence
import requests

from django import template as django_template
from os.path import relpath
from camomilla.settings import (
    REGISTERED_TEMPLATES_APPS,
    INTEGRATIONS_ASTRO_ENABLE,
    INTEGRATIONS_ASTRO_URL,
)


def get_templates(request=None) -> Sequence[str]:
    files = []

    for engine in django_template.loader.engines.all():

        if REGISTERED_TEMPLATES_APPS:
            dirs = [
                d
                for d in engine.template_dirs
                if any(app in str(d) for app in REGISTERED_TEMPLATES_APPS)
            ]
        else:
            # Exclude pip installed site package template dirs
            dirs = [
                d
                for d in engine.template_dirs
                if "site-packages" not in str(d) or "camomilla" in str(d)
            ]

        for d in dirs:
            base = Path(d)
            files.extend(relpath(f, d) for f in base.rglob("*.html"))

    if INTEGRATIONS_ASTRO_ENABLE and request is not None:
        try:
            response = requests.get(
                INTEGRATIONS_ASTRO_URL + "/api/templates",
                cookies={
                    "sessionid": request.COOKIES.get("sessionid"),
                    "csrftoken": request.COOKIES.get("csrftoken"),
                },
            )
            if response.status_code == 200:
                astro_templates = response.json()
                for template in astro_templates:
                    files.append(template)
        except:
            pass

    return files



from django.urls import path
from django.shortcuts import redirect


url_patterns = [
    path(
        "profiles/me/", lambda _: redirect("../../users/current/"), name="profiles-me"
    ),
    path("sitemap/", lambda _: redirect("../pages/"), name="sitemap"),
]



from django.utils.module_loading import import_string
from django.conf import settings

from django import VERSION as DJANGO_VERSION


def get_default_storage_class():
    if DJANGO_VERSION >= (4, 2):
        storage = settings.STORAGES["default"]["BACKEND"]
    else:
        storage = settings.DEFAULT_FILE_STORAGE
    return import_string(storage)



import os
import mock

from django.http import Http404
from django.test import TestCase
from django.test import RequestFactory
from django.utils.translation import activate, get_language

from camomilla.models import Page, Article
from camomilla.templatetags.camomilla_filters import filter_content, alternate_urls

from requests import RequestException


class CamomillaFiltersTestCase(TestCase):
    def setUp(self):
        pass

    def test_filter_content(self):
        Page.objects.create(
            identifier="path", title="Path", permalink="/path", status="PUB"
        )
        request_factory = RequestFactory()
        request = request_factory.get("/path")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request)
        content = filter_content(page, "content1")
        self.assertEqual(content.identifier, "content1")
        self.assertEqual(content.content, "")
        content.content = "Hello World!"
        content.save()
        page = Page.get(request)
        content = filter_content(page, "content1")
        self.assertEqual(content.identifier, "content1")
        self.assertEqual(content.content, "Hello World!")

    def test_filter_alternate_urls(self):
        Page.objects.create(
            identifier="path", title="Path", permalink="/path", status="PUB"
        )
        request = RequestFactory().get("/path")
        request.META["HTTP_HOST"] = "localhost"
        page = Page.get(request)
        alt_urls = dict(alternate_urls(page, request))
        self.assertEqual(alt_urls, {"it": None})



import os
from tests.fixtures import load_asset
from django.conf import settings


def load_asset_and_remove_media(filename):
    asset = load_asset(filename)
    if os.path.exists(f"{settings.MEDIA_ROOT}/{filename}"):
        os.remove(f"{settings.MEDIA_ROOT}/{filename}")
    return asset



from functools import cached_property
from typing import Iterable, List
from modeltranslation.translator import NotRegistered, translator
from modeltranslation.utils import build_localized_fieldname
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from camomilla.utils.getters import pointed_getter
from camomilla.utils.translation import is_translatable
from camomilla.utils.translation import nest_to_plain, plain_to_nest
from camomilla.settings import API_TRANSLATION_ACCESSOR, LANGUAGE_CODES


class TranslationsMixin(serializers.ModelSerializer):
    """
    This mixin adds support for modeltranslation fields.
    It automatically nests all translations fields (es. title_en) under a "translations" field.

    This means that, in representation, the serializer will transform:
    `{"title_en": "Hello", "title_it": "Ciao"}` -> `{"translations": {"en": {"title": "Hello"}, "it": {"title": "Ciao"}}`

    While in deserialization, the serializer will transform:
    `{"translations": {"en": {"title": "Hello"}, "it": {"title": "Ciao"}}` -> `{"title_en": "Hello", "title_it": "Ciao"}`
    """

    def _transform_input(self, data):
        return nest_to_plain(
            data, self.translation_fields or [], API_TRANSLATION_ACCESSOR
        )

    def _transform_output(self, data):
        return plain_to_nest(
            data, self.translation_fields or [], API_TRANSLATION_ACCESSOR
        )

    @cached_property
    def translation_fields(self) -> List[str]:
        try:
            return translator.get_options_for_model(self.Meta.model).get_field_names()
        except NotRegistered:
            return []

    @property
    def _writable_fields(self) -> Iterable[serializers.Field]:
        for field in super()._writable_fields:
            if field.field_name not in self.translation_fields:
                yield field

    def to_internal_value(self, data):
        return super().to_internal_value(self._transform_input(data))

    def to_representation(self, instance):
        return self._transform_output(super().to_representation(instance))

    def run_validation(self, *args, **kwargs):
        try:
            return super().run_validation(*args, **kwargs)
        except ValidationError as ex:
            ex.detail.update(self._transform_input(ex.detail))
            raise ValidationError(detail=ex.detail)

    @property
    def is_translatable(self):
        return is_translatable(pointed_getter(self, "Meta.model"))


class RemoveTranslationsMixin(serializers.ModelSerializer):
    """
    This mixin removes all translations fields (es. title_en) from the serializer.
    It's useful when you want to create a serializer that doesn't need to include all translations fields.

    If request is passed in context, this serializer becomes aware of the query parameter "included_translations".
    If the value is "all", all translations fields are included.
    If the value is a comma separated list of languages (es. "en,it"), only the specified translations fields are included.
    """

    @cached_property
    def translation_fields(self):
        try:
            return translator.get_options_for_model(self.Meta.model).get_field_names()
        except NotRegistered:
            return []

    def get_default_field_names(self, declared_fields, model_info):
        request = self.context.get("request", False)
        included_translations = request and request.GET.get(
            "included_translations", False
        )
        if included_translations == "all":
            return super().get_default_field_names(declared_fields, model_info)
        elif included_translations is not False:
            included_translations = included_translations.split(",")
        else:
            included_translations = []

        field_names = super().get_default_field_names(declared_fields, model_info)
        for lang in LANGUAGE_CODES:
            if lang not in included_translations:
                for field in self.translation_fields:
                    localized_fieldname = build_localized_fieldname(field, lang)
                    if localized_fieldname in field_names:
                        field_names.remove(localized_fieldname)
        return field_names



from .fields import FieldsOverrideMixin
from .filter_fields import FilterFieldsMixin
from .json import JSONFieldPatchMixin
from .language import LangInfoMixin
from .nesting import NestMixin
from .optimize import SetupEagerLoadingMixin
from .ordering import OrderingMixin
from .page import AbstractPageMixin
from .translation import TranslationsMixin, RemoveTranslationsMixin


__all__ = [
    "FieldsOverrideMixin",
    "FilterFieldsMixin",
    "JSONFieldPatchMixin",
    "LangInfoMixin",
    "NestMixin",
    "SetupEagerLoadingMixin",
    "OrderingMixin",
    "AbstractPageMixin",
    "TranslationsMixin",
    "RemoveTranslationsMixin",
]



from camomilla.utils.normalization import dict_merge

def test_dict_merge_simple():
    a = {'x': 1}
    b = {'y': 2}
    result = dict_merge(a.copy(), b)
    assert result == {'x': 1, 'y': 2}

def test_dict_merge_overwrite():
    a = {'x': 1}
    b = {'x': 2}
    result = dict_merge(a.copy(), b)
    assert result == {'x': 2}

def test_dict_merge_nested():
    a = {'x': {'y': 1}}
    b = {'x': {'z': 2}}
    result = dict_merge(a.copy(), b)
    assert result == {'x': {'y': 1, 'z': 2}}

def test_dict_merge_nested_overwrite():
    a = {'x': {'y': 1}}
    b = {'x': {'y': 2}}
    result = dict_merge(a.copy(), b)
    assert result == {'x': {'y': 2}}

def test_dict_merge_empty():
    a = {}
    b = {'foo': 1}
    result = dict_merge(a.copy(), b)
    assert result == {'foo': 1}

def test_dict_merge_both_empty():
    a = {}
    b = {}
    result = dict_merge(a.copy(), b)
    assert result == {}



from django import template
from django.utils.safestring import mark_safe
from django.forms.models import model_to_dict
from structured.pydantic.models import BaseModel
import json
import re
import html
from datetime import datetime
from uuid import UUID

register = template.Library()


def custom_json_serializer(obj):
    """Serialize custom objects to JSON."""
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


def pretty_dict(data, indent_level=0):
    """
    Recursive function to format a dictionary into a pretty string.
    Args:
        data (dict): The dictionary to format.
        indent_level (int): The current indentation level.
    Returns:
        str: A pretty-printed string representation of the dictionary.
    """
    indent = "  " * indent_level
    result = []

    for key, value in data.items():

        if isinstance(value, dict):
            result.append(f"{indent}'{key}': {{")
            result.append(pretty_dict(value, indent_level + 1))
            result.append(f"{indent}}},")

        elif isinstance(value, list):
            result.append(f"{indent}'{key}': [")
            for item in value:
                if isinstance(item, dict):
                    result.append(pretty_dict(item, indent_level + 1))
                else:
                    result.append(
                        f"{indent}  {json.dumps(item, default=custom_json_serializer)},"
                    )
            result.append(f"{indent}],")

        else:
            result.append(
                f"{indent}'{key}': {json.dumps(value, default=custom_json_serializer)},"
            )

    return "\n".join(result).rstrip(",")


@register.filter
def to_pretty_dict(instance):
    data = model_to_dict(instance)

    for key, value in data.items():
        if isinstance(value, BaseModel):
            data[key] = value.model_dump()
        elif isinstance(value, UUID):
            data[key] = str(value)

    formatted_dict = "{\n" + pretty_dict(data, indent_level=1) + "\n}"

    escaped = html.escape(formatted_dict)

    # This to highlight the keys in the JSON
    highlighted = re.sub(
        r"(&#x27;)([^&#]+?)(&#x27;):",
        r"<span style='color:#df3079'>'\2'</span>:",
        escaped,
    )

    return mark_safe(f"<pre>{highlighted}</pre>")



from django.db.models.query import QuerySet
from django.core.exceptions import ObjectDoesNotExist
from django.apps import apps
from django.db import models
from django.utils import timezone
from django.db.utils import ProgrammingError, OperationalError
from typing import Sequence, Tuple

URL_NODE_RELATED_NAME = "%(app_label)s_%(class)s"


class PageQuerySet(QuerySet):

    __UrlNodeModel = None

    @property
    def UrlNodeModel(self):
        if not self.__UrlNodeModel:
            self.__UrlNodeModel = apps.get_model("camomilla", "UrlNode")
        return self.__UrlNodeModel

    def get_permalink_kwargs(self, kwargs):
        return list(
            set(kwargs.keys()).intersection(
                set(self.UrlNodeModel.LANG_PERMALINK_FIELDS + ["permalink"])
            )
        )

    def get(self, *args, **kwargs):
        permalink_args = self.get_permalink_kwargs(kwargs)
        if len(permalink_args):
            try:
                node = self.UrlNodeModel.objects.get(
                    **{arg: kwargs.pop(arg) for arg in permalink_args}
                )
                kwargs["url_node"] = node
            except ObjectDoesNotExist:
                raise self.model.DoesNotExist(
                    "%s matching query does not exist." % self.model._meta.object_name
                )
        return super(PageQuerySet, self).get(*args, **kwargs)


class UrlNodeManager(models.Manager):

    def get_reverse_pages_relations(self):
        """
        Get all reverse relations coming from AbstractPages models.
        This is used to annotate the UrlNode with the related page fields.
        """
        from camomilla.models.page import AbstractPage

        relations = []

        for field in self.model._meta.get_fields():
            if not (hasattr(field, "related_model") and field.one_to_one):
                continue

            if not issubclass(field.related_model, AbstractPage):
                continue

            if field.remote_field.name != "url_node":
                continue

            related_name = field.get_accessor_name()
            relations.append(
                {
                    "name": related_name,
                    "model": field.related_model,
                    "field_name": field.remote_field.name,
                    "field": field,
                }
            )
        return relations

    @property
    def related_names(self):
        self._related_names = getattr(self, "_related_names", None)
        if self._related_names is None:
            self._related_names = list(
                set([rel["name"] for rel in self.get_reverse_pages_relations()])
            )
        return self._related_names

    def _annotate_fields(
        self,
        qs: models.QuerySet,
        field_names: Sequence[Tuple[str, models.Field, models.Value]],
    ):
        for field_name, output_field, default in field_names:
            whens = [
                models.When(
                    related_name=related_name,
                    then=models.F("__".join([related_name, field_name])),
                )
                for related_name in self.related_names
            ]
            qs = qs.annotate(
                **{
                    field_name: models.Case(
                        *whens, output_field=output_field, default=default
                    )
                }
            )
        return self._annotate_is_public(qs)

    def _annotate_is_public(self, qs: models.QuerySet):
        return qs.annotate(
            is_public=models.Case(
                models.When(status="PUB", then=True),
                models.When(
                    status="PLA", publication_date__lte=timezone.now(), then=True
                ),
                default=False,
                output_field=models.BooleanField(default=False),
            )
        )

    def get_queryset(self):
        try:
            return self._annotate_fields(
                super().get_queryset(),
                [
                    (
                        "indexable",
                        models.BooleanField(),
                        models.Value(None, models.BooleanField()),
                    ),
                    (
                        "status",
                        models.CharField(),
                        models.Value("DRF", models.CharField()),
                    ),
                    (
                        "publication_date",
                        models.DateTimeField(),
                        models.Value(timezone.now(), models.DateTimeField()),
                    ),
                    (
                        "date_updated_at",
                        models.DateTimeField(),
                        models.Value(timezone.now(), models.DateTimeField()),
                    ),
                ],
            )
        except (ProgrammingError, OperationalError):
            return super().get_queryset()



import io
import json
import pytest
from django.conf import settings
from django.test import RequestFactory
from django.core.files.uploadedfile import SimpleUploadedFile
from camomilla.parsers import MultipartJsonParser

@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_multipart_json_parser_parses_json_and_files(monkeypatch):
    # Prepare multipart data
    json_data = json.dumps({"foo": "bar", "nested": {"baz": 1}, "nested_list": [{"qux": "quux"}, {"corge": "grault"}]})
    file_content = b"filecontent"
    upload = SimpleUploadedFile("test.txt", file_content, content_type="text/plain")
    
    # Simulate multipart POST
    factory = RequestFactory()
    data = {
        "data": json_data,
        "nested.file": upload,
        "nested_list.1.file": upload,
    }
    request = factory.post("/", data)
    request.upload_handlers = []
    request.META["CONTENT_TYPE"] = "multipart/form-data; boundary=BoUnDaRy"
    
    # Patch DjangoMultiPartParser to return our data
    class DummyParser:
        def __init__(self, *a, **kw): pass
        def parse(self):
            return ( {"data": json_data}, {"nested.file": upload, "nested_list.1.file": upload,},)
    monkeypatch.setattr("camomilla.parsers.DjangoMultiPartParser", DummyParser)

    parser = MultipartJsonParser()
    parsed = parser.parse(io.BytesIO(b""), "multipart/form-data", {"request": request})
    assert parsed["foo"] == "bar"
    assert parsed["nested"]["baz"] == 1
    assert isinstance(parsed["nested"]["file"], SimpleUploadedFile)
    assert isinstance(parsed["nested_list"][1]["file"], SimpleUploadedFile)
    parsed["nested"]["file"].seek(0)
    assert parsed["nested"]["file"].read() == file_content
    parsed["nested_list"][1]["file"].seek(0)
    assert parsed["nested_list"][1]["file"].read() == file_content

@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_multipart_json_parser_handles_parse_error(monkeypatch):
    factory = RequestFactory()
    request = factory.post("/", {})
    request.upload_handlers = []
    request.META["CONTENT_TYPE"] = "multipart/form-data; boundary=BoUnDaRy"
    
    class DummyParser:
        def __init__(self, *a, **kw): pass
        def parse(self):
            from django.http.multipartparser import MultiPartParserError
            raise MultiPartParserError("fail")
    monkeypatch.setattr("camomilla.parsers.DjangoMultiPartParser", DummyParser)
    parser = MultipartJsonParser()
    from rest_framework.exceptions import ParseError
    with pytest.raises(ParseError) as exc:
        parser.parse(io.BytesIO(b""), "multipart/form-data", {"request": request})
    assert "Multipart form parse error" in str(exc.value)



import functools
from django.utils.translation import activate
from django.views.decorators.cache import cache_page
from camomilla.settings import LANGUAGE_CODES, DEFAULT_LANGUAGE


def active_lang(*args, **kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if len(args) and hasattr(args[0], "request"):
                request = args[0].request
            else:
                request = args[0] if len(args) else kwargs.get("request", None)
            lang = DEFAULT_LANGUAGE
            if request and hasattr(request, "GET"):
                lang = request.GET.get("lang", request.GET.get("language", lang))
            if request and hasattr(request, "data"):
                lang = request.data.pop("lang", request.data.pop("language", lang))
            if lang and lang in LANGUAGE_CODES:
                activate(lang)
                request.LANGUAGE_CODE = lang
            return func(*args, **kwargs)

        return wrapped_func

    return decorator


def staff_excluded_cache(timing=None):
    def decorator(func):
        if timing is None:
            return func  # No caching applied

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            request = args[0] if len(args) else kwargs.get("request", None)
            if request and hasattr(request, "user"):
                user = request.user
                if user.is_authenticated and user.is_staff:
                    return func(*args, **kwargs)
            return cache_page(timing)(func)(*args, **kwargs)

        return wrapped_func

    return decorator



from .bulk_actions import BulkDeleteMixin
from .language import GetUserLanguageMixin
from .optimize import OptimViewMixin
from .ordering import OrderingMixin
from .pagination import PaginateStackMixin, TrigramSearchMixin
from .permissions import CamomillaBasePermissionMixin


__all__ = [
    "BulkDeleteMixin",
    "GetUserLanguageMixin",
    "OptimViewMixin",
    "OrderingMixin",
    "PaginateStackMixin",
    "TrigramSearchMixin",
    "CamomillaBasePermissionMixin",
]



import pytest
import json
from django.test import TransactionTestCase
from camomilla.models import Media
from .utils.api import login_superuser
from .utils.media import load_asset_and_remove_media
from rest_framework.test import APIClient

client = APIClient()


class MediaTestCase(TransactionTestCase):
    reset_sequences = True

    def setUp(self):
        self.client = APIClient()
        token = login_superuser()
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)

    def test_media_api_crud(self):
        # Create media 1
        asset = load_asset_and_remove_media("10595073.png")
        response = self.client.post(
            "/api/camomilla/media/",
            {
                "file": asset,
                "data": json.dumps(
                    {
                        "translations": {
                            "en": {
                                "alt_text": "Test 1",
                                "title": "Test 1",
                                "description": "Test 1",
                            }
                        }
                    }
                ),
            },
            format="multipart",
        )
        assert response.status_code == 201
        assert Media.objects.count() == 1
        media = Media.objects.first()
        assert media.alt_text == "Test 1"
        assert media.title == "Test 1"
        assert media.description == "Test 1"
        assert media.file.name == "10595073.png"

        # Create media 2
        asset = load_asset_and_remove_media("37059501.png")
        response = self.client.post(
            "/api/camomilla/media/",
            {
                "file": asset,
                "data": json.dumps(
                    {
                        "translations": {
                            "en": {
                                "alt_text": "Test 2",
                                "title": "Test 2",
                                "description": "Test 2",
                            }
                        }
                    }
                ),
            },
            format="multipart",
        )
        assert response.status_code == 201
        assert Media.objects.count() == 2
        media = Media.objects.first()  # Ordering in model is descending -pk
        assert media.alt_text == "Test 2"
        assert media.title == "Test 2"
        assert media.description == "Test 2"
        assert media.file.name == "37059501.png"

        # Read media
        response = self.client.get("/api/camomilla/media/2/")
        assert response.status_code == 200
        assert response.json()["id"] == 2
        assert response.json()["title"] == "Test 2"
        assert response.json()["file"] == "http://testserver/media/37059501.png"

        # Read medias
        response = self.client.get("/api/camomilla/media/")
        assert response.status_code == 200
        assert response.json()[0]["id"] == 2  # Ordering in model is descending -pk
        assert response.json()[0]["title"] == "Test 2"
        assert response.json()[1]["id"] == 1
        assert response.json()[1]["title"] == "Test 1"

        # Delete media
        response = self.client.delete("/api/camomilla/media/2/")
        assert response.status_code == 204
        assert len(Media.objects.all()) == 1
        media = Media.objects.last()
        assert media.id == 1
        assert media.title == "Test 1"

    def test_media_compression(self):
        asset = load_asset_and_remove_media("Sample-jpg-image-10mb.jpg")
        asset_size = asset.size
        response = self.client.post(
            "/api/camomilla/media/",
            {
                "file": asset,
                "data": json.dumps(
                    {
                        "translations": {
                            "en": {
                                "alt_text": "Test",
                                "title": "Test",
                                "description": "Test",
                            }
                        }
                    }
                ),
            },
            format="multipart",
        )
        assert response.status_code == 201
        assert Media.objects.count() == 1
        media = Media.objects.first()
        assert media.file.size < asset_size
        assert media.file.size < 1000000  # 1MB

    def test_inflating_prevent(self):
        asset = load_asset_and_remove_media("optimized.jpg")
        asset_size = asset.size
        response = self.client.post(
            "/api/camomilla/media/",
            {
                "file": asset,
                "data": json.dumps(
                    {
                        "translations": {
                            "en": {
                                "alt_text": "Test",
                                "title": "Test",
                                "description": "Test",
                            }
                        }
                    }
                ),
            },
            format="multipart",
        )
        assert response.status_code == 201
        assert Media.objects.count() == 1
        media = Media.objects.first()
        assert media.file.size < asset_size



import pytest
from django.test import TransactionTestCase
from rest_framework.test import APIClient
from .utils.api import login_superuser
from example.website.models import TestModel


class ModelAPiRegisterTestCase(TransactionTestCase):
    reset_sequences = True
    
    def setUp(self):
        self.client = APIClient()
        token = login_superuser()
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token)

    def test_model_api_register_access(self):
        client = APIClient()
        response = client.post("/api/models/test-model/")
        assert response.status_code == 401

    def test_model_api_register_crud(self):
        # Create test model 1
        response = self.client.post(
            "/api/models/test-model/", {"title": "title_test_model_1"}, format="json"
        )
        assert response.status_code == 201
        assert len(TestModel.objects.all()) == 1
        test_model = TestModel.objects.first()
        assert test_model.id == 1
        assert test_model.title == "title_test_model_1"

        # Create test model 2
        response = self.client.post(
            "/api/models/test-model/", {"title": "title_test_model_2"}, format="json"
        )
        assert response.status_code == 201
        assert len(TestModel.objects.all()) == 2
        test_model = TestModel.objects.last()
        assert test_model.id == 2
        assert test_model.title == "title_test_model_2"

        # Update test model 2
        response = self.client.patch(
            "/api/models/test-model/2/",
            {
                "title": "title_test_model_2_updated",
            },
            format="json",
        )
        assert response.status_code == 200
        assert len(TestModel.objects.all()) == 2
        test_model = TestModel.objects.last()
        assert test_model.id == 2
        assert test_model.title == "title_test_model_2_updated"

        # Read test model 2
        response = self.client.get("/api/models/test-model/2/")
        assert response.status_code == 200
        assert response.json()["id"] == 2
        assert response.json()["title"] == "title_test_model_2_updated"

        # Read test models
        response = self.client.get("/api/models/test-model/")
        assert response.status_code == 200
        assert response.json()[0]["id"] == 2
        assert response.json()[0]["title"] == "title_test_model_2_updated"
        assert response.json()[1]["id"] == 1
        assert response.json()[1]["title"] == "title_test_model_1"

        # Delete page
        response = self.client.delete("/api/models/test-model/2/")
        assert response.status_code == 204
        assert len(TestModel.objects.all()) == 1
        test_model = TestModel.objects.last()
        assert test_model.id == 1
        assert test_model.title == "title_test_model_1"

    def test_model_api_register_listing(self):
        # Create test model 1
        response = self.client.post(
            "/api/models/test-model/", {"title": "title_test_model_1"}, format="json"
        )
        assert response.status_code == 201

        # Create test model 2
        response = self.client.post(
            "/api/models/test-model/", {"title": "title_test_model_2"}, format="json"
        )
        assert response.status_code == 201

        # Create test model 3
        response = self.client.post(
            "/api/models/test-model/", {"title": "title_test_model_3"}, format="json"
        )
        assert response.status_code == 201

        # Simple Response
        response = self.client.get("/api/models/test-model/")
        assert response.status_code == 200
        assert len(response.json()) == 3
        assert response.json()[0]["id"] == 3
        assert response.json()[1]["id"] == 2
        assert response.json()[2]["id"] == 1

        # Pagination
        response = self.client.get("/api/models/test-model/?items=2")
        assert response.status_code == 200
        assert len(response.json()["items"]) == 2
        assert response.json()["items"][0]["id"] == 3
        assert response.json()["items"][1]["id"] == 2
        assert response.json()["paginator"]["count"] == 3
        assert response.json()["paginator"]["page"] == 1
        assert response.json()["paginator"]["has_next"] == True
        assert response.json()["paginator"]["has_previous"] == False
        assert response.json()["paginator"]["pages"] == 2
        assert response.json()["paginator"]["page_size"] == 2

        response = self.client.get("/api/models/test-model/?items=2&page=2")
        assert response.status_code == 200
        assert len(response.json()["items"]) == 1
        assert response.json()["items"][0]["id"] == 1
        assert response.json()["paginator"]["count"] == 3
        assert response.json()["paginator"]["page"] == 2
        assert response.json()["paginator"]["has_next"] == False
        assert response.json()["paginator"]["has_previous"] == True
        assert response.json()["paginator"]["pages"] == 2
        assert response.json()["paginator"]["page_size"] == 2

        # Filtering
        response = self.client.get(
            "/api/models/test-model/?fltr=title=title_test_model_2"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1
        assert response.json()[0]["id"] == 2
        assert response.json()[0]["title"] == "title_test_model_2"

        response = self.client.get("/api/models/test-model/?fltr=title=not_real_title")
        assert response.status_code == 200
        assert len(response.json()) == 0

        response = self.client.get(
            "/api/models/test-model/?fltr=title=title_test_model_2&fltr=id=2"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1
        assert response.json()[0]["id"] == 2
        assert response.json()[0]["title"] == "title_test_model_2"

        response = self.client.get(
            "/api/models/test-model/?fltr=title=title_test_model_2&fltr=id=3"
        )
        assert response.status_code == 200
        assert len(response.json()) == 0

        response = self.client.get(
            "/api/models/test-model/?fltr=title__in=[title_test_model_2,title_test_model_3]"
        )
        assert response.status_code == 200
        assert len(response.json()) == 2
        assert response.json()[0]["id"] == 3
        assert response.json()[0]["title"] == "title_test_model_3"
        assert response.json()[1]["id"] == 2
        assert response.json()[1]["title"] == "title_test_model_2"

        response = self.client.get(
            "/api/models/test-model/?fltr=title__in=[title_test_model_4,title_test_model_3]"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1
        assert response.json()[0]["id"] == 3
        assert response.json()[0]["title"] == "title_test_model_3"

    def test_model_api_register_listing_filtered_model(self):
        # Create filter argument register model 1
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "test 1"},
            format="json",
        )
        assert response.status_code == 201

        # Create filter argument register model 2
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "pippo 2"},
            format="json",
        )
        assert response.status_code == 201

        # Create filter argument register model 3
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "3 test"},
            format="json",
        )
        assert response.status_code == 201

        response = self.client.get("/api/models/filtered-register-model/")
        assert response.status_code == 200
        assert len(response.json()) == 2
        assert response.json()[0]["id"] == 3
        assert response.json()[0]["field_filtered"] == "3 test"
        assert response.json()[1]["id"] == 1
        assert response.json()[1]["field_filtered"] == "test 1"

    def test_model_api_register_base_view_search(self):
        # Verify search_fields = ["description"]
        # Create custom arguments register model 1
        response = self.client.post(
            "/api/models/custom-base-arguments-register-model/",
            {"description": "description_1"},
            format="json",
        )
        assert response.status_code == 201

        # Create custom arguments register model 2
        response = self.client.post(
            "/api/models/custom-base-arguments-register-model/",
            {"description": "description_2"},
            format="json",
        )
        assert response.status_code == 201

        # Create custom arguments register model 3
        response = self.client.post(
            "/api/models/custom-base-arguments-register-model/",
            {"description": "description_3"},
            format="json",
        )
        assert response.status_code == 201

        response = self.client.get(
            "/api/models/custom-base-arguments-register-model/?search=not_real_description"
        )
        assert response.status_code == 200
        assert len(response.json()) == 0

        response = self.client.get(
            "/api/models/custom-base-arguments-register-model/?search=description_1"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1
        assert response.json()[0]["id"] == 1
        assert response.json()[0]["description"] == "description_1"

    def test_model_api_register_base_serializer_search(self):
        # Verify description = BaseModelSerializer.CharField(min_length=3)
        # Create custom arguments register model 1
        response = self.client.post(
            "/api/models/custom-base-arguments-register-model/",
            {"description": "12"},
            format="json",
        )
        assert response.status_code == 400

        # Create custom arguments register model 2
        response = self.client.post(
            "/api/models/custom-base-arguments-register-model/",
            {"description": "1234"},
            format="json",
        )
        assert response.status_code == 201

    def test_model_api_register_view_search(self):
        # Verify {"search_fields": ["name"]}
        # Create custom arguments register model 1
        response = self.client.post(
            "/api/models/custom-arguments-register-model/",
            {"name": "name_1"},
            format="json",
        )
        assert response.status_code == 201

        # Create custom arguments register model 2
        response = self.client.post(
            "/api/models/custom-arguments-register-model/",
            {"name": "name_2"},
            format="json",
        )
        assert response.status_code == 201

        # Create custom arguments register model 3
        response = self.client.post(
            "/api/models/custom-arguments-register-model/",
            {"name": "name_3"},
            format="json",
        )
        assert response.status_code == 201

        response = self.client.get(
            "/api/models/custom-arguments-register-model/?search=not_real_name"
        )
        assert response.status_code == 200
        assert len(response.json()) == 0

        response = self.client.get(
            "/api/models/custom-arguments-register-model/?search=name_1"
        )
        assert response.status_code == 200
        assert len(response.json()) == 1
        assert response.json()[0]["name"] == "name_1"

    def test_model_api_register_serializer_search(self):
        # Verify {"fields": ["name"]}
        # Create custom arguments register model
        response = self.client.post(
            "/api/models/custom-arguments-register-model/",
            {"name": "1234"},
            format="json",
        )
        assert response.status_code == 201
        assert "id" not in response.json()
        assert response.json()["name"] == "1234"

    def test_model_api_register_listing_filtered_model(self):
        # Verify filters={"field_filtered__icontains": "test"}
        # Create filter argument register model 1
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "test 1"},
            format="json",
        )
        assert response.status_code == 201

        # Create filter argument register model 2
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "pippo 2"},
            format="json",
        )
        assert response.status_code == 201

        # Create filter argument register model 3
        response = self.client.post(
            "/api/models/filtered-register-model/",
            {"field_filtered": "3 test"},
            format="json",
        )
        assert response.status_code == 201

        response = self.client.get("/api/models/filtered-register-model/")
        assert response.status_code == 200
        assert len(response.json()) == 2
        assert response.json()[0]["id"] == 3
        assert response.json()[0]["field_filtered"] == "3 test"
        assert response.json()[1]["id"] == 1
        assert response.json()[1]["field_filtered"] == "test 1"



import responses
from django.test import TestCase
from camomilla.models import Page
from camomilla.utils.templates import get_templates
from camomilla.theme.admin import PageAdmin
from django.contrib.admin.sites import AdminSite
from camomilla.settings import INTEGRATIONS_ASTRO_URL


class MockRequest:
    pass


request = MockRequest()


class MockRequestWithCookies:
    def __init__(self):
        self.COOKIES = {"sessionid": "mock_session_id", "csrftoken": "mock_csrf_token"}


request_with_cookies = MockRequestWithCookies()


class AdminPageFormTestCase(TestCase):
    def setUp(self):
        self.astro_api_url = INTEGRATIONS_ASTRO_URL + "/api/templates"

    @responses.activate
    def test_admin_page_form_astro_error(self):
        responses.add(
            responses.GET,
            self.astro_api_url,
            status=500,  # Simulating a server error
        )

        page_admin = PageAdmin(Page, AdminSite())
        form = page_admin.get_form(request)()
        self.assertEqual(len(list(form.fields)), 35)
        self.assertTrue("template" in list(form.fields))
        self.assertListEqual(
            form.fields["template"].widget.choices,
            [("", "---------")] + [(t, t) for t in get_templates(request)],
        )

    @responses.activate
    def test_admin_page_form(self):
        responses.add(
            responses.GET,
            self.astro_api_url,
            json=["mock_template/1", "mock_template/2"],
            status=200,
        )

        page_admin = PageAdmin(Page, AdminSite())
        form = page_admin.get_form(request_with_cookies)()
        self.assertEqual(len(list(form.fields)), 35)
        self.assertTrue("template" in list(form.fields))
        self.assertListEqual(
            form.fields["template"].widget.choices,
            [("", "---------")] + [(t, t) for t in get_templates(request_with_cookies)],
        )
        self.assertEqual(responses.calls[0].request.url, self.astro_api_url)



from camomilla.models import Page
from camomilla.models.page import UrlNode, UrlRedirect
from camomilla.serializers import PageSerializer
from camomilla.serializers.page import RouteSerializer
from camomilla.utils.translation import url_lang_decompose
from camomilla.views.base import BaseModelViewset
from camomilla.views.decorators import staff_excluded_cache
from camomilla.views.mixins import BulkDeleteMixin, GetUserLanguageMixin
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response
from rest_framework import permissions
from django.utils.translation.trans_real import activate as activate_language
from django.shortcuts import get_object_or_404
from camomilla.settings import PAGE_ROUTER_CACHE


class PageViewSet(GetUserLanguageMixin, BulkDeleteMixin, BaseModelViewset):
    queryset = Page.objects.all()
    serializer_class = PageSerializer
    model = Page


@api_view(["GET"])
@staff_excluded_cache(PAGE_ROUTER_CACHE)
@permission_classes(
    [
        permissions.AllowAny,
    ]
)
def pages_router(request, permalink=""):
    redirect = UrlRedirect.find_redirect_from_url(f"/{permalink}")
    if redirect:
        redirect = redirect.redirect()
        return Response({"redirect": redirect.url, "status": redirect.status_code})
    url_decomposition = url_lang_decompose(permalink)
    if not url_decomposition["permalink"].startswith("/"):
        url_decomposition["permalink"] = f"/{url_decomposition['permalink']}"
    activate_language(url_decomposition["language"])
    node: UrlNode = get_object_or_404(UrlNode, permalink=url_decomposition["permalink"])
    return Response(RouteSerializer(node, context={"request": request}).data)



import pytest
from rest_framework.test import APIClient
from .utils.api import login_user, login_superuser, login_staff

client = APIClient()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_right_permissions():
    response = client.post("/api/models/test-model/", {"title": "test"}, format="json")
    assert response.status_code == 401
    response = client.get("/api/models/test-model/")
    assert response.status_code == 401
    token = login_user()
    client.credentials(HTTP_AUTHORIZATION="Token " + token)
    response = client.post("/api/models/test-model/", {"title": "test"}, format="json")
    assert response.status_code == 403
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    token = login_staff()
    client.credentials(HTTP_AUTHORIZATION="Token " + token)
    response = client.post("/api/models/test-model/", {"title": "test"}, format="json")
    assert response.status_code == 403
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    token = login_superuser()
    client.credentials(HTTP_AUTHORIZATION="Token " + token)
    response = client.post("/api/models/test-model/", {"title": "test"}, format="json")
    assert response.status_code == 201
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    assert len(response.json()) == 1
    response = client.get("/api/models/test-model/1/")
    assert response.status_code == 200
    response = client.patch(
        "/api/models/test-model/1/", {"title": "updated"}, format="json"
    )
    assert response.status_code == 200
    assert response.json()["title"] == "updated"
    response = client.delete("/api/models/test-model/1/")
    assert response.status_code == 204
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    assert len(response.json()) == 0
    response = client.get("/api/models/test-model/1/")
    assert response.status_code == 404
    assert response.json() in [
        {"detail": "Not found."},
        {"detail": "No TestModel matches the given query."},
    ]



from django.contrib.auth.models import User
from rest_framework.test import APIClient

client = APIClient()


def login_superuser():
    User.objects.create_superuser("admin", "admin@test.com", "adminadmin")
    response = client.post(
        "/api/camomilla/token-auth/", {"username": "admin", "password": "adminadmin"}
    )
    return response.json()["token"]


def login_user():
    User.objects.create_user("user", "user@test.com", "useruser")
    response = client.post(
        "/api/camomilla/token-auth/", {"username": "user", "password": "useruser"}
    )
    return response.json()["token"]


def login_staff():
    User.objects.create_user("staff", "staff@test.com", "staffstaff", is_staff=True)
    response = client.post(
        "/api/camomilla/token-auth/", {"username": "staff", "password": "staffstaff"}
    )
    return response.json()["token"]



from django.db import connections
from datetime import datetime
from django.utils import dateparse

_mysql_datetime_patched = False  # module level flag to avoid double patching


def monkey_patch_mysql_datetime():  # pragma: no cover - environment/driver specific
    """Normalize MySQL DATETIME values that arrive as strings.

    In some environments (notably with certain PyMySQL versions / settings)
    MySQL returns DATETIME columns as raw strings. Django's MySQL backend then
    hands those straight to its timezone utilities expecting a datetime.
    That raises AttributeError on value.utcoffset(). We monkeypatch the
    backend converter once to coerce any string into a datetime before the
    original logic runs.
    """
    global _mysql_datetime_patched
    if _mysql_datetime_patched:
        return
    try:
        for alias in connections:
            conn = connections[alias]
            if conn.vendor != "mysql":
                continue
            ops = conn.ops
            original = ops.convert_datetimefield_value

            def _patched_convert(value, expression, connection, _orig=original):
                if isinstance(value, str):
                    dt = dateparse.parse_datetime(value.replace(" ", "T", 1)) or None
                    if dt is None:
                        try:
                            dt = datetime.fromisoformat(value)
                        except Exception:
                            dt = None
                    if dt is not None:
                        value = dt
                return _orig(value, expression, connection)

            ops.convert_datetimefield_value = _patched_convert  # type: ignore[attr-defined]
            _mysql_datetime_patched = True
    except Exception:
        # Fail silently; if conversion still fails tests will reveal it.
        pass



from camomilla.utils.getters import safe_getter, pointed_getter, find_and_replace_dict

def test_safe_getter_dict():
    d = {'a': 1}
    assert safe_getter(d, 'a') == 1
    assert safe_getter(d, 'b', 42) == 42

def test_safe_getter_object():
    class Dummy:
        foo = 'bar'
    obj = Dummy()
    assert safe_getter(obj, 'foo') == 'bar'
    assert safe_getter(obj, 'baz', 99) == 99

def test_pointed_getter_dict():
    d = {'a': {'b': {'c': 5}}}
    assert pointed_getter(d, 'a.b.c') == 5
    assert pointed_getter(d, 'a.b.x', 'nope') == 'nope'

def test_pointed_getter_object():
    class Dummy:
        pass
    obj = Dummy()
    obj.child = Dummy()
    obj.child.value = 123
    assert pointed_getter(obj, 'child.value') == 123
    assert pointed_getter(obj, 'child.missing', 'default') == 'default'

def test_find_and_replace_dict_simple():
    d = {'a': 1, 'b': 2}
    def pred(key, value):
        return value * 2
    result = find_and_replace_dict(d, pred)
    assert result == {'a': 2, 'b': 4}

def test_find_and_replace_dict_nested():
    d = {'a': {'b': 2}, 'c': 3}
    def pred(key, value):
        return value if not isinstance(value, int) else value + 1
    result = find_and_replace_dict(d, pred)
    assert result == {'a': {'b': 3}, 'c': 4}



from django.test import TestCase
from rest_framework.test import APIClient
from .utils.api import login_superuser
from example.website.models import (
    ExposedRelatedPageModel,
    UnexposedRelatedPageModel,
    RelatedPageModel,
)

client = APIClient()


class PageRelationApi(TestCase):
    def setUp(self):
        token = login_superuser()
        client.credentials(HTTP_AUTHORIZATION="Token " + token)

    def test_exposed_relation(self):
        exposed_related_page_model = ExposedRelatedPageModel.objects.create(
            title="ExposedRelatedPageModel 1",
            permalink="exposed-related-page-model-1",
            status="PUB",
            autopermalink=False,
        )
        related_page_model = RelatedPageModel.objects.create(
            title="RelatedPageModel 1",
            permalink="related-page-model-1",
            status="PUB",
            autopermalink=False,
        )
        related_page_model.exposed_pages.add(exposed_related_page_model)
        related_page_model.save()

        response = client.get("/api/camomilla/pages-router/related-page-model-1")
        assert response.status_code == 200
        data = response.json()
        assert data["exposed_pages"][0]["id"] == exposed_related_page_model.id

        response = client.get(
            "/api/camomilla/pages-router/exposed-related-page-model-1"
        )
        assert response.status_code == 200
        data = response.json()
        assert (
            "related_pages" in data
        ), "Exposed related pages should be included in the API response"

    def test_unexposed_relation(self):
        unexposed_related_page_model = UnexposedRelatedPageModel.objects.create(
            title="UnexposedRelatedPageModel 1",
            permalink="unexposed-related-page-model-1",
            status="PUB",
            autopermalink=False,
        )
        related_page_model = RelatedPageModel.objects.create(
            title="RelatedPageModel 1",
            permalink="related-page-model-1",
            status="PUB",
            autopermalink=False,
        )

        related_page_model.unexposed_pages.add(unexposed_related_page_model)
        related_page_model.save()

        response = client.get("/api/camomilla/pages-router/related-page-model-1")
        assert response.status_code == 200
        data = response.json()
        assert data["unexposed_pages"][0]["id"] == unexposed_related_page_model.id

        response = client.get(
            "/api/camomilla/pages-router/unexposed-related-page-model-1"
        )
        assert response.status_code == 200
        data = response.json()
        assert (
            "related_pages" not in data
        ), "Unexposed related pages should not be included in the API response"



from rest_framework import serializers
from camomilla.models import UrlNode
from camomilla.serializers.validators import UniquePermalinkValidator
from typing import TYPE_CHECKING
from structured.contrib.restframework import StructuredModelSerializer


if TYPE_CHECKING:
    from camomilla.models.page import AbstractPage


class AbstractPageMixin(StructuredModelSerializer, serializers.ModelSerializer):
    """
    This mixin is needed to serialize AbstractPage models.
    It provides permalink validation and some extra fields serialization.

    Use it as a base class for your serializer if you need to serialize custom AbstractPage models.
    """

    breadcrumbs = serializers.SerializerMethodField()
    routerlink = serializers.CharField(read_only=True)
    template_file = serializers.SerializerMethodField()

    def get_template_file(self, instance: "AbstractPage"):
        return instance.get_template_path()

    def get_breadcrumbs(self, instance: "AbstractPage"):
        return instance.breadcrumbs

    @property
    def translation_fields(self):
        return super().translation_fields + ["permalink"]

    def get_default_field_names(self, *args):
        from camomilla.serializers.mixins.translation import RemoveTranslationsMixin

        default_fields = super().get_default_field_names(*args)
        filtered_fields = getattr(self, "filtered_fields", [])
        if len(filtered_fields) > 0:
            return filtered_fields
        if RemoveTranslationsMixin in self.__class__.__bases__:  # noqa: E501
            return default_fields
        return list(
            set(
                [f for f in default_fields if f != "url_node"]
                + UrlNode.LANG_PERMALINK_FIELDS
                + ["permalink"]
            )
        )

    def build_field(self, field_name, info, model_class, nested_depth):
        if field_name in UrlNode.LANG_PERMALINK_FIELDS + ["permalink"]:
            return serializers.CharField, {
                "required": False,
                "allow_blank": True,
            }
        return super().build_field(field_name, info, model_class, nested_depth)

    def get_validators(self):
        return super().get_validators() + [UniquePermalinkValidator()]



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import Http404
from django.shortcuts import get_object_or_404
from rest_framework.decorators import action
from rest_framework.response import Response

from camomilla.models import AbstractPage, Menu
from camomilla.models.page import UrlNode
from camomilla.serializers import ContentTypeSerializer, MenuSerializer
from camomilla.serializers.page import UrlNodeSerializer
from camomilla.views.base import BaseModelViewset
from camomilla.views.decorators import active_lang


class MenuViewSet(BaseModelViewset):
    queryset = Menu.objects.all()
    serializer_class = MenuSerializer
    model = Menu

    lookup_field = "key"

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        assert lookup_url_kwarg in self.kwargs, (
            "Expected view %s to be called with a URL keyword argument "
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            "attribute on the view correctly."
            % (self.__class__.__name__, lookup_url_kwarg)
        )
        filter = Q(**{self.lookup_field: self.kwargs[lookup_url_kwarg]})
        if (
            isinstance(self.kwargs[lookup_url_kwarg], str)
            and self.kwargs[lookup_url_kwarg].isnumeric()
            or isinstance(self.kwargs[lookup_url_kwarg], int)
        ):
            filter |= Q(pk=self.kwargs[lookup_url_kwarg])

        obj = get_object_or_404(queryset, filter)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    @action(detail=False, methods=["get"], url_path="page_types")
    def page_types(self, request, *args, **kwargs):
        return Response(
            ContentTypeSerializer(
                [
                    model
                    for model in ContentType.objects.all()
                    if model.model_class()
                    and issubclass(model.model_class(), AbstractPage)
                ],
                many=True,
            ).data
        )

    @action(detail=False, methods=["get"], url_path=r"page_types/(?P<content_id>\w+)")
    def page_type_instances(self, request, content_id, *args, **kwargs):
        content_type = ContentType.objects.filter(pk=content_id).first()
        if not content_type or not issubclass(content_type.model_class(), AbstractPage):
            raise Http404("No object matches the given query.")
        return Response(
            [
                {
                    "id": obj.pk,
                    "name": str(obj),
                    "url_node_id": obj.url_node.pk,
                    "model": f"{content_type.app_label}.{content_type.model}",
                }
                for obj in content_type.model_class().objects.exclude(
                    url_node__isnull=True
                )
            ]
        )

    @active_lang()
    @action(detail=False, methods=["get"], url_path="search_urlnode")
    def search_urlnode(self, request, *args, **kwargs):
        url_node = request.GET.get("q", "")
        qs = UrlNode.objects.filter(permalink__icontains=url_node).order_by("permalink")
        return Response(UrlNodeSerializer(qs, many=True).data)



import json
import os
from django.core.files.uploadedfile import SimpleUploadedFile


def load_json_fixture(filename):
    with open(os.path.join(os.path.dirname(__file__), "json", filename), "r") as f:
        return json.load(f)


def load_asset(filename):
    with open(os.path.join(os.path.dirname(__file__), "assets", filename), "rb") as f:
        up_file = SimpleUploadedFile(filename, f.read())
        return up_file



from camomilla.serializers.fields.related import RelatedField
from camomilla.serializers.utils import build_standard_model_serializer
from camomilla import settings


class NestMixin:
    """
    This mixin automatically creates nested serializers for relational fields.
    The depth of the nesting can be set using the "depth" attribute of the Meta class.
    If the depth is not set, the serializer will use the value coming from the settings.

    CAMOMILLA = { "API": {"NESTING_DEPTH": 10} }
    """

    def __init__(self, *args, **kwargs):
        self._depth = kwargs.pop("depth", None)
        return super().__init__(*args, **kwargs)

    def build_nested_field(self, field_name, relation_info, nested_depth):
        return self.build_relational_field(field_name, relation_info, nested_depth)

    def build_relational_field(
        self, field_name, relation_info, nested_depth=settings.API_NESTING_DEPTH + 1
    ):
        nested_depth = nested_depth if self._depth is None else self._depth
        field_class, field_kwargs = super().build_relational_field(
            field_name, relation_info
        )
        if (
            field_class is RelatedField and nested_depth > 1
        ):  # stop recursion one step before the jump :P
            field_kwargs["serializer"] = build_standard_model_serializer(
                relation_info[1], nested_depth - 1
            )
        return field_class, field_kwargs



import pytest
from camomilla.utils import setters

def test_set_key_dict():
    d = {'a': 1}
    result = setters.set_key(d, 'b', 2)
    assert result['b'] == 2
    assert d['b'] == 2

def test_set_key_list():
    l = [1, 2]
    result = setters.set_key(l, 1, 3)
    assert result[1] == 3
    result = setters.set_key(l, 2, 4)
    assert result[2] == 4

def test_set_key_object():
    class Dummy:
        pass
    obj = Dummy()
    setters.set_key(obj, 'foo', 'bar')
    assert obj.foo == 'bar'

def test_get_key_list():
    l = [10, 20]
    assert setters.get_key(l, 1, 'x') == 20
    assert setters.get_key(l, 5, 'x') == 'x'

def test_pointed_setter_dict():
    d = {'a': {'b': 1}}
    setters.pointed_setter(d, 'a.b', 2)
    assert d['a']['b'] == 2

def test_pointed_setter_list():
    l = [[0, 1], [2, 3]]
    setters.pointed_setter(l, '1.1', 99)
    assert l[1][1] == 99

def test_set_key_list_append():
    l = [1]
    # key out of range, should append
    setters.set_key(l, 2, 42)
    assert l == [1, 42]

def test_set_key_invalid_type():
    class Dummy:
        pass
    obj = Dummy()
    # Should set attribute if not dict or list
    setters.set_key(obj, 'bar', 123)
    assert obj.bar == 123

def test_get_key_invalid_index():
    l = [1, 2]
    # Out of range index returns default
    assert setters.get_key(l, 10, 'default') == 'default'

def test_pointed_setter_new_path():
    d = {}
    # Should create nested dicts
    setters.pointed_setter(d, 'foo.bar.baz', 7)
    assert d['foo']['bar']['baz'] == 7



import pytest
from rest_framework.test import APIClient
from .fixtures import load_json_fixture
from .utils.api import login_superuser
from example.website.models import SimpleRelationModel


client = APIClient()


@pytest.fixture(autouse=True)
def init_test():
    token = login_superuser()
    client.credentials(HTTP_AUTHORIZATION="Token " + token)
    SimpleRelationModel.objects.bulk_create(
        [SimpleRelationModel(name=f"test{i}") for i in range(1, 10)]
    )


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_simple_relation_model_api_endpoint():
    response = client.get("/api/models/simple-relation-model/")
    assert response.status_code == 200
    assert len(response.json()) == 9
    response = client.get("/api/models/simple-relation-model/1/")
    assert response.status_code == 200
    assert response.json()["name"] == "test1"
    response = client.patch(
        "/api/models/simple-relation-model/1/", {"name": "updated"}, format="json"
    )
    assert response.status_code == 200
    assert response.json()["name"] == "updated"
    response = client.delete("/api/models/simple-relation-model/1/")
    assert response.status_code == 204
    response = client.get("/api/models/simple-relation-model/")
    assert response.status_code == 200
    assert len(response.json()) == 8
    response = client.get("/api/models/simple-relation-model/1/")
    assert response.status_code == 404
    assert response.json() in [
        {"detail": "Not found."},
        {"detail": "No SimpleRelationModel matches the given query."},
    ]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_test_model_api_endpoint():
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    assert response.json() == []
    test_model_data = load_json_fixture("test-model-api.json")
    response = client.post("/api/models/test-model/", test_model_data, format="json")
    assert response.status_code == 201
    assert response.json()["title"] == test_model_data["title"]
    assert (
        response.json()["structured_data"]["name"]
        == test_model_data["structured_data"]["name"]
    )
    assert (
        response.json()["structured_data"]["age"]
        == test_model_data["structured_data"]["age"]
    )
    assert (
        response.json()["structured_data"]["child"]["name"]
        == test_model_data["structured_data"]["child"]["name"]
    )
    assert (
        response.json()["structured_data"]["childs"][0]["name"]
        == test_model_data["structured_data"]["childs"][0]["name"]
    )
    assert (
        response.json()["structured_data"]["fk_field"]["id"]
        == test_model_data["structured_data"]["fk_field"]["id"]
    )
    assert (
        response.json()["structured_data"]["qs_field"][0]["id"]
        == test_model_data["structured_data"]["qs_field"][0]["id"]
    )
    response = client.get("/api/models/test-model/")
    assert response.status_code == 200
    assert len(response.json()) == 1
    response = client.get("/api/models/test-model/1/")
    assert response.status_code == 200
    assert response.json()["title"] == test_model_data["title"]
    assert (
        response.json()["structured_data"]["name"]
        == test_model_data["structured_data"]["name"]
    )
    assert (
        response.json()["structured_data"]["age"]
        == test_model_data["structured_data"]["age"]
    )
    assert (
        response.json()["structured_data"]["child"]["name"]
        == test_model_data["structured_data"]["child"]["name"]
    )
    assert (
        response.json()["structured_data"]["childs"][0]["name"]
        == test_model_data["structured_data"]["childs"][0]["name"]
    )
    assert (
        response.json()["structured_data"]["fk_field"]["id"]
        == test_model_data["structured_data"]["fk_field"]["id"]
    )
    assert (
        response.json()["structured_data"]["qs_field"][0]["id"]
        == test_model_data["structured_data"]["qs_field"][0]["id"]
    )
    response = client.patch(
        "/api/models/test-model/1/", {"title": "updated"}, format="json"
    )
    assert response.status_code == 200
    assert response.json()["title"] == "updated"



import pytest
from django.db.models import Q
from camomilla.utils.query_parser import (
    ConditionParser,
)


@pytest.mark.parametrize(
    "query, expected_q",
    [
        # Single condition
        ("name__icontains=foo", Q(name__icontains="foo")),
        # Multiple conditions with AND
        (
            "name__icontains=foo AND age__gt=21",
            Q(name__icontains="foo") & Q(age__gt=21),
        ),
        # Multiple conditions with OR
        (
            "name__icontains=foo OR name__icontains=bar",
            Q(name__icontains="foo") | Q(name__icontains="bar"),
        ),
        # Mixed AND and OR conditions
        (
            "name__icontains=foo OR (age__gt=21 AND city__iexact='New York')",
            Q(name__icontains="foo") | (Q(age__gt=21) & Q(city__iexact="New York")),
        ),
        # Nested parentheses with OR inside AND
        (
            "((name__icontains=foo) OR (name__icontains=bar)) AND name__icontains=baz",
            (Q(name__icontains="foo") | Q(name__icontains="bar"))
            & Q(name__icontains="baz"),
        ),
        # Complex nested conditions with multiple OR and AND
        (
            "((name__icontains=foo) OR (name__icontains=bar) OR (name__icontains=buz)) AND age__gt=25 AND city__iexact='LA'",
            (
                Q(name__icontains="foo")
                | Q(name__icontains="bar")
                | Q(name__icontains="buz")
            )
            & Q(age__gt=25)
            & Q(city__iexact="LA"),
        ),
        # Single condition with no parentheses
        ("age__lte=30", Q(age__lte=30)),
        # Simple nested OR and AND with more levels of nesting
        (
            "(name__icontains=foo AND (age__gt=21 OR city__iexact='New York')) OR country__iexact='US'",
            (Q(name__icontains="foo") & (Q(age__gt=21) | Q(city__iexact="New York")))
            | Q(country__iexact="US"),
        ),
    ],
)
def test_condition_parser(query, expected_q):
    parser = ConditionParser(query)
    q_object = parser.parse_to_q()
    assert q_object.__str__() == expected_q.__str__()
